#include <sys/types.h>
#include <sys/param.h>

#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/ioccom.h>
#include <sys/kmem.h>
#include <sys/lkm.h>
#include <sys/namei.h>

#include <machine/int_fmtio.h>

#include "netbsd-dm.h"
#include "dm.h"

extern struct dm_softc *dm_sc;

static int dm_dbg_print_flags(int);

static int
dm_dbg_print_flags(int flags)
{
	printf("dbg_print --- %d\n",flags);
	
	if (flags & DM_READONLY_FLAG)
		printf("dbg_flags: DM_READONLY_FLAG set In/Out\n");

	if (flags & DM_SUSPEND_FLAG)
		printf("dbg_flags: DM_SUSPEND_FLAG set In/Out \n");

	if (flags & DM_PERSISTENT_DEV_FLAG)
		printf("dbg_flags: DM_PERSISTENT_DEV_FLAG set In\n");

	if (flags & DM_STATUS_TABLE_FLAG)
		printf("dbg_flags: DM_STATUS_TABLE_FLAG set In\n");

	if (flags & DM_ACTIVE_PRESENT_FLAG)
		printf("dbg_flags: DM_ACTIVE_PRESENT_FLAG set Out\n");

	if (flags & DM_INACTIVE_PRESENT_FLAG)
		printf("dbg_flags:  DM_INACTIVE_PRESENT_FLAG set Out\n");

	if (flags & DM_BUFFER_FULL_FLAG)
		printf("dbg_flags: DM_BUFFER_FULL_FLAG set Out\n");

	if (flags & DM_SKIP_BDGET_FLAG)
		printf("dbg_flags: DM_SKIP_BDGET_FLAG set In\n");

	if (flags & DM_SKIP_LOCKFS_FLAG)
		printf("dbg_flags: DM_SKIP_LOCKFS_FLAG set In\n");

	if (flags & DM_NOFLUSH_FLAG)
		printf("dbg_flags: DM_NOFLUSH_FLAG set In\n");
	
	return 0;
}

/* Get version ioctl call I do it as default therefore this function is unused now. */
int
dm_get_version_ioctl(prop_dictionary_t dm_dict)
{
	int r;

	r = 0;

	printf("dm_get_version_ioctl called\n");

	return 0;
}

/* List targets and their versions. */
int
dm_list_versions_ioctl(prop_dictionary_t dm_dict)
{
	prop_array_t target_list;
	
	int r;
	uint32_t flags;
	
	r = 0;

	printf("dm_list_versions_ioctl called\n");

	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	
	dm_dbg_print_flags(flags);

	target_list = dm_target_prop_list();

	prop_dictionary_set(dm_dict,DM_IOCTL_CMD_DATA,target_list);
	
	return r;
}

/*
 * Create in-kernel entry for device. Device atributes such as name, uuid are
 * taken from dictionary.
 */
int
dm_dev_create_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;
	char *name, *uuid;
	int r,flags;
	
	r = 0;
	flags = 0;
	name = NULL;	
	uuid = NULL;
	
	/* Get needed values from dictionary. */
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	
	dm_dbg_print_flags(flags);

	/* Lookup name and uuid if device already exist quit. */
	if ((dm_dev_lookup_name(name) != NULL) &&
	    (dm_dev_lookup_uuid(uuid) != NULL))
		return EEXIST;

	if ((dmv = dm_dev_alloc()) == NULL)
		return ENOMEM;

	if (uuid)
		memcpy(dmv->uuid,uuid,sizeof(dmv->uuid));

	if (name)
		strlcpy(dmv->name,name,DM_NAME_LEN);

	dmv->flags = flags;

	dmv->minor = ++(dm_sc->sc_minor_num);

	dmv->ref_cnt = 0;
	dmv->event_nr = 0;
	dmv->cur_active_table = 0;

	/* Initialize tables. */
	SLIST_INIT(&dmv->tables[0]);
	SLIST_INIT(&dmv->tables[1]);

	/* init device list of physical devices. */
	SLIST_INIT(&dmv->pdevs);
	/*
	 * Locking strategy here is this: this is per device mutex lock
	 * and it should be taken when we work with device. Almost all
	 * ioctl callbacks for tables and devices must acquire this lock.
     	 */

	prop_dictionary_set_uint64(dm_dict,DM_IOCTL_DEV,dmv->minor);
	
	mutex_init(&dmv->dev_mtx,MUTEX_DEFAULT,IPL_NONE);

	/* Test readonly flag change anything only if it is not set*/
	if (flags & DM_READONLY_FLAG)
		dm_dev_free(dmv);
	else
		r = dm_dev_insert(dmv);
	
	return r;
}

int
dm_dev_list_ioctl(prop_dictionary_t dm_dict)
{
	prop_array_t dev_list;
	
	int r;
	uint32_t flags;
	
	r = 0;

	printf("dm_dev_list called\n");

	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	
	dm_dbg_print_flags(flags);
	
	dev_list = dm_dev_prop_list();

	prop_dictionary_set(dm_dict,DM_IOCTL_CMD_DATA,dev_list);
	
	return 0;
}

/*
 * Rename selected devices old name is in struct dm_ioctl.
 * newname is taken from dictionary
 *
 * <key>cmd_data</key>
 *  <array>
 *   <string>...</string>
 *  </array>
 */
int
dm_dev_rename_ioctl(prop_dictionary_t dm_dict)
{
	prop_array_t cmd_array;
	struct dm_dev *dmv;
	
	char *name,*uuid,*n_name;
	uint32_t flags;
	
	/* Get needed values from dictionary. */
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	cmd_array = prop_dictionary_get(dm_dict,DM_IOCTL_CMD_DATA);

	prop_array_get_cstring_nocopy(cmd_array,0,(const char **)&n_name);
	
	if (strlen(n_name) + 1 > DM_NAME_LEN)
		return EINVAL;
	
	if (((dmv = dm_dev_lookup_uuid(uuid)) == NULL) ||
	    (dmv = dm_dev_lookup_name(name)) == NULL)
		return ENOENT;
	
	/* Test readonly flag change anything only if it is not set*/
	if (!(flags & DM_READONLY_FLAG))
		strlcpy(dmv->name,n_name,DM_NAME_LEN);
	
	return 0;
}

int
dm_dev_remove_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;

	char *name, *uuid;
	uint32_t flags;

	/* Get needed values from dictionary. */
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	
	dm_dbg_print_flags(flags);
	
	if (((dmv = dm_dev_lookup_uuid(uuid)) == NULL) &&
	    (dmv = dm_dev_lookup_name(name)) == NULL)
		return ENOENT;

	if (!(flags & DM_READONLY_FLAG)) {
		if (dmv->ref_cnt == 0){
			dm_table_destroy(&dmv->tables[dmv->cur_active_table]);

			/* Remove all pdevs from this device if they are unused
			   close vnode and remove them from global pdev list, too. */
			dm_pdev_decr(&dmv->pdevs);
	
			/* Test readonly flag change anything only if it is not set*/
	
			dm_dev_rem(name); /* XXX. try to use uuid, too */
		}
		
	}
	return 0;
}

int
dm_dev_status_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;
	char *name, *uuid;

	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	
	if (((dmv = dm_dev_lookup_name(name)) == NULL) &&
	    ((dmv = dm_dev_lookup_uuid(uuid)) == NULL))
		return 0;

	prop_dictionary_set_uint32(dm_dict,DM_IOCTL_OPEN,dmv->ref_cnt);
	prop_dictionary_set_uint32(dm_dict,DM_IOCTL_FLAGS,dmv->flags);
	
	prop_dictionary_set_uint64(dm_dict,DM_IOCTL_DEV,dmv->minor);
	
	return 0;
}

int
dm_dev_suspend_ioctl(prop_dictionary_t dm_dict)
{
	return 0;
}

int
dm_dev_resume_ioctl(prop_dictionary_t dm_dict)
{
	return 0;
}


int
dm_table_clear_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;
	struct dm_table  *tbl;

	char *name, *uuid;
	int flags;

	dmv = NULL;

	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);
	
	printf("Clearing inactive table from device: %s--%s\n",name,uuid);
	
	if (((dmv = dm_dev_lookup_name(name)) == NULL) &&
	    ((dmv = dm_dev_lookup_uuid(uuid)) == NULL))
		return ENOENT;

	/* Select unused table */
	tbl = &dmv->tables[1-dmv->cur_active_table]; 

	/* Test readonly flag change anything only if it is not set*/
	if (!(flags & DM_READONLY_FLAG))
			dm_table_destroy(tbl);

	return 0;
}

/* XXX Dunno what we should do here. */
int
dm_table_deps_ioctl(prop_dictionary_t dm_dict)
{
	return 0;
}

/* Load new device mapper table to device  */

int
dm_table_load_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;
	struct dm_pdev *dmp;
	struct dm_table_entry *table_en, *last_table;
	struct dm_table  *tbl;
	struct dm_target *target;

	prop_object_iterator_t iter;
	prop_array_t cmd_array;
	prop_dictionary_t target_dict;
	
	char *name, *uuid, *type, *str, *str_ptr;
	int flags;

	char *dev_name;

	flags = 0;
	name = NULL;
	uuid = NULL;
	dmv = NULL;
	dmp = NULL;
	last_table = NULL;
	str_ptr = NULL;

	char *xml;
	
	xml = prop_dictionary_externalize(dm_dict);
	printf("%s\n",xml);
	
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);

	cmd_array = prop_dictionary_get(dm_dict,DM_IOCTL_CMD_DATA);
	iter = prop_array_iterator(cmd_array);
	
	dm_dbg_print_flags(flags);	

	printf("Loading table to device: %s--%s\n",name,uuid);
	
	if (((dmv = dm_dev_lookup_name(name)) == NULL) &&
	    ((dmv = dm_dev_lookup_uuid(uuid)) == NULL))
		return ENOENT;

	/* Select unused table */
	tbl = &dmv->tables[1-dmv->cur_active_table]; 

	/*
	 * I have to check if this table slot is not used by another table list.
	 * if it is used I should free them.
	 */

	dm_table_destroy(tbl);
	
	while((target_dict = prop_object_iterator_next(iter)) != NULL){

		prop_dictionary_get_cstring_nocopy(target_dict,DM_TABLE_TYPE,(const char **)&type);
		
		//printf("Table sector_start %016" PRIu64 ", length %016" PRIu64 ", target %s\n",
		//   dm_target_spec->sector_start, dm_target_spec->length,
		//   dm_target_spec->target_type);
		
		/* If we want to deny table with 2 or more different target we should do it here */
		if ((target = dm_target_lookup_name(type)) == NULL)
			return ENOENT;
		
		if ((table_en=kmem_alloc(sizeof(struct dm_table_entry),KM_NOSLEEP)) == NULL)
			return ENOMEM;
		
		prop_dictionary_get_uint64(target_dict,DM_TABLE_START,&table_en->start);
		prop_dictionary_get_uint64(target_dict,DM_TABLE_LENGTH,&table_en->length);

		table_en->target=target;
		table_en->dm_dev=dmv;
		
		/*
		 * There is a parameter string after dm_target_spec
		 * structure which  points to /dev/wd0a 284 part of
		 * table. String str points to this text.
		 */

		prop_dictionary_get_cstring_nocopy(target_dict,
		    DM_TABLE_PARAMS,(const char **)&str);

		if (strlen(str) != 0) {

			str_ptr = str; /* save pointer to start of string */
			
			printf("trg_str: %s\n",str);
			
		        dev_name = strsep(&str," ");

			printf("namei: %s, str: %s\n",dev_name,str);
						
			if ((dmp = dm_pdev_incr(dev_name)) == NULL) {
				kmem_free(table_en,sizeof(struct dm_table_entry));
				return ENOENT;
			}
			
			if (!(flags & DM_READONLY_FLAG)){
				dm_pdev_insert(dmp); /* insert pdev to global list */
						
				/* insert pdev to device list  */
				SLIST_INSERT_HEAD(&dmv->pdevs,dmp,next_pdev); 
			}
		}
		
		/* Test readonly flag change anything only if it is not set*/
		if (!(flags & DM_READONLY_FLAG)){

			if (SLIST_EMPTY(tbl))
				/* insert this table to head */
				SLIST_INSERT_HEAD(tbl,table_en,next);
			else
				SLIST_INSERT_AFTER(last_table,table_en,next);

			prop_dictionary_get_cstring(target_dict,
			    DM_TABLE_PARAMS,(char **)&table_en->params);

			/*
			 * XXX I have to calculate argc
			 */
			if (target->init != NULL)
				target->init(dmv,&table_en->target_config,2,&str_ptr);

			last_table = table_en;
			
		} else {
			kmem_free(table_en,sizeof(struct dm_table_entry));
			dm_pdev_destroy(dmp);
		}
			
	}
	
	dmv->cur_active_table = 1 - dmv->cur_active_table;

	prop_object_release(cmd_array);

	return 0;
}

int
dm_table_status_ioctl(prop_dictionary_t dm_dict)
{
	struct dm_dev *dmv;
	struct dm_table  *tbl;
	struct dm_table_entry *table_en;

	prop_array_t cmd_array;
	prop_dictionary_t target_dict;
	
	uint32_t rec_size;

	char *name, *uuid;
	int flags,i;
	
	dmv = NULL;
	uuid = NULL;
	name = NULL;
	flags = 0;
	rec_size = 0;
	i=0;
	
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_NAME,(const char **)&name);
	prop_dictionary_get_cstring_nocopy(dm_dict,DM_IOCTL_UUID,(const char **)&uuid);
	prop_dictionary_get_uint32(dm_dict,DM_IOCTL_FLAGS,&flags);

	cmd_array = prop_dictionary_get(dm_dict,DM_IOCTL_CMD_DATA);
	
	printf("Status of device tables: %s--%s\n",name,uuid);
	
	if (((dmv = dm_dev_lookup_name(name)) == NULL) &&
	    ((dmv = dm_dev_lookup_uuid(uuid)) == NULL))
		return ENOENT;

	tbl = &dmv->tables[dmv->cur_active_table];
	
	SLIST_FOREACH(table_en,tbl,next)
	{
		target_dict = prop_dictionary_create();
		printf("%016" PRIu64 ", length %016" PRIu64 ", target %s\n",
		    table_en->start,table_en->length,table_en->target->name);

		prop_dictionary_set_uint64(target_dict,DM_TABLE_START,table_en->start);
		prop_dictionary_set_uint64(target_dict,DM_TABLE_LENGTH,table_en->length);

		prop_dictionary_set_cstring(target_dict,DM_TABLE_TYPE,table_en->target->name);
		prop_dictionary_set_cstring(target_dict,DM_TABLE_PARAMS,table_en->params);
		
		prop_dictionary_set_int32(target_dict,DM_TABLE_STAT,dmv->cur_active_table);

		prop_array_set(cmd_array,i,target_dict);

		prop_object_release(target_dict);
		
		i++;
	}

//	prop_object_release(cmd_array);

	return 0;
}


int
dm_check_version(prop_dictionary_t dm_dict)
{
	size_t i;
	int version[3];
	prop_array_t ver;
	int r;

	r = 0;
	
	ver = prop_dictionary_get(dm_dict,DM_IOCTL_VERSION);
	
	for(i=0;i<3;i++) {
		prop_array_get_uint32(ver,i,&version[i]);
		printf("Check version %d\n",version[i]);
	}

	if (DM_VERSION_MAJOR != version[0] || DM_VERSION_MINOR < version[1])
	{
		printf("libdevmapper/kernel version mismatch "
		    "kernel: %d.%d.%d libdevmapper: %d.%d.%d\n",
		    DM_VERSION_MAJOR,DM_VERSION_MINOR,DM_VERSION_PATCHLEVEL,
		    version[0],version[1],version[2]);
		r = EIO;
		goto out;
	}

	prop_array_set_uint32(ver,0,DM_VERSION_MAJOR);
	prop_array_set_uint32(ver,1,DM_VERSION_MINOR);
	prop_array_set_uint32(ver,2,DM_VERSION_PATCHLEVEL);

out:	
	return r;
}
