

#include <sys/types.h>
#include <sys/param.h>

#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/ioccom.h>
#include <sys/kmem.h>
#include <sys/queue.h>

#include <sys/namei.h>
#include <sys/vnode.h>

#include "dm.h"

static struct dm_pdev *dm_pdev_alloc(char *);

/*
 * I should use rwlock here.
 */
kmutex_t dm_pdev_mutex;

/*
 *Search for name in SLIST and return apropriate pointer.
 */
struct dm_pdev*
dm_pdev_lookup_name(char *dm_pdev_name)
{
	struct dm_pdev *dm_pdev;
	int dlen; int slen;

	slen = strlen(dm_pdev_name)+1;

	mutex_enter(&dm_pdev_mutex);

	SLIST_FOREACH(dm_pdev, &dm_pdev_list, next_pdev) {
		dlen = strlen(dm_pdev->name)+1;

		if (slen != dlen)
			continue;
		
		if (strncmp(dm_pdev_name,dm_pdev->name,slen) == 0){
			mutex_exit(&dm_pdev_mutex);
			return dm_pdev;
		}
	}
	mutex_exit(&dm_pdev_mutex);

	return NULL;
}

/*
 * Search for dev_name in list of used physical devices
 */
struct dm_pdev*
dm_pdev_incr(char *dev_name)
{
	struct dm_pdev *dmp;
	struct nameidata nd;
	int error;
	
	dmp = dm_pdev_lookup_name(dev_name);

	if (dmp != NULL) {
		dmp->ref_cnt++;
		return NULL;
	}

	if ((dmp = dm_pdev_alloc(dev_name)) == NULL)
		return NULL;

	NDINIT(&nd, LOOKUP, FOLLOW, UIO_SYSSPACE, dev_name);
	
	error = namei(&nd);
	
	if (error) {
		printf("namei error\n");
		dm_pdev_destroy(dmp);
		return NULL;
	}
	
	error = VOP_OPEN(nd.ni_vp, FREAD|FWRITE, FSCRED);
	
	if (error) {
		printf("VOP_OPEN error\n");
		dm_pdev_destroy(dmp);
		return NULL;
	}
	
	dmp->pdev_vnode = nd.ni_vp;
	dmp->ref_cnt = 1;
	
	return dmp;
}

/*
 * Initialize pdev subsystem.
 */
int
dm_pdev_init()
{
	
	SLIST_INIT(&dm_pdev_list); /* initialize global pdev list */
	
	mutex_init(&dm_pdev_mutex,MUTEX_DEFAULT,IPL_NONE);

	return 0;
}

/*
 * Insert new pdev struct into the SLIST. 
 */
int
dm_pdev_insert(struct dm_pdev *dm_pdev)
{
	struct dm_pdev *dmp;
	
	dmp = dm_pdev_lookup_name(dm_pdev->name);

	if (dmp != NULL)
		return 0;
	
	mutex_enter(&dm_pdev_mutex);

	SLIST_INSERT_HEAD(&dm_pdev_list, dm_pdev, next_pdev);

	mutex_exit(&dm_pdev_mutex);

	return 0;
}

/*
 * Allocat new pdev structure if is not already present and
 * set name.
 */
static struct dm_pdev*
dm_pdev_alloc(char *name)
{
	struct dm_pdev *dmp;

	if ((dmp = kmem_alloc(sizeof(struct dm_pdev)+strlen(name)+1,KM_NOSLEEP))
	    == NULL)
		return NULL;

	strlcpy(dmp->name,name,MAX_DEV_NAME);
	
	dmp->ref_cnt = 0;
	dmp->pdev_vnode = NULL;
	
	return dmp;
}

/*
 * Destroy allocated dm_pdev.
 */
int
dm_pdev_destroy(struct dm_pdev *dmp)
{

	if (dmp == NULL)
		return ENOENT;

	if (dmp->pdev_vnode != NULL)
		VOP_CLOSE(dmp->pdev_vnode,FREAD|FWRITE,FSCRED);
	
	kmem_free(dmp,sizeof(*dmp));

	dmp = NULL;
	
	return 0;
}

/*
 * This funcion is called from dm_dev_remove_ioctl.
 * When I'm removing device from list, I have to decrement
 * reference counter. If reference counter is 0 I will remove
 * dmp from global list and from device list to. And I will CLOSE
 * dmp vnode too.
 */

/*
 * I'm guarding only global list with mutex, device should be
 * locked elsewhere.
 */

int
dm_pdev_decr(struct dm_pdevs *head)
{
	struct dm_pdev *dmp;
	if (head == NULL)
		return ENOENT;

	
	SLIST_FOREACH(dmp,head,next_pdev) {

		dmp->ref_cnt--;

		SLIST_REMOVE(head,dmp,dm_pdev,next_pdev);
		
		/* If this was last reference remove dmp from global list also. */
		if (dmp->ref_cnt == 0) {
			mutex_enter(&dm_pdev_mutex);
			SLIST_REMOVE(&dm_pdev_list,dmp, dm_pdev,next_pdev); 
			dm_pdev_destroy(dmp);
			mutex_exit(&dm_pdev_mutex);
		}
	}

	return 0;
}

